
local tablet = AceLibrary("Tablet-2.0")
local ae = AceLibrary("AceEvent-2.0")
local timetable

local maxttitems, sortedframes, hooks, events, names, frameshooked = 50, {}, {}, {}, {}
local events, insort = {}, {}
local blacklist = {
	["AceEvent20Frame"] = true,
}


tekSupport_Events = tekSupport:NewModule("Event Handlers")


local function callandsave(frame, event, mem, time, ...)
	mem, time = collectgarbage("count") - mem, GetTime() - time

	if mem >= 0 then
		if not events[event] then events[event] = {} end
		if not events[event][frame] then
			events[event][frame] = {event = event, name = names[frame], count = 0, time = 0, mem = 0}
			table.insert(sortedframes, events[event][frame])
		end
		local e = events[event][frame]
		e.time = e.time + time
		e.mem = e.mem + mem
		e.count = e.count + 1
	end

	return ...
end


local function HookedFrame(frame, event, ...)
	local mem, time, f = collectgarbage("count"), GetTime(), hooks[frame]
	if not f then return end
	return callandsave(frame, event, mem, time, f(frame, event, ...))
end


function tekSupport_Events:OnEnable()
	ae:EnableDebugging()
	timetable = ae.debugTable

	if frameshooked then return end

	local f = EnumerateFrames()
	while f do
		local script = f and f:GetScript("OnEvent")
		if script then
			local name = f.GetName and f:GetName() or f.name or tostring(f)
			if not blacklist[name] then
				names[f] = name
				hooks[f] = script
				f:SetScript("OnEvent", HookedFrame)
			end
		end
		f = EnumerateFrames(f)
	end
	frameshooked = true
end


local function SortMem(a, b)
	if a.count == 0 then return end
	if b.count == 0 then return true end
	local dif = (a.mem or 0)/a.count - (b.mem or 0)/b.count
	if dif ~= 0 then return dif > 0 end
	if (names[a] or a.name) ~= (names[b] or b.name) then return (names[a] or a.name) < (names[b] or b.name) end
	return (events[a] or a.event) < (events[b] or b.event)
end


local function SortTime(a, b)
	if a.count == 0 then return end
	if b.count == 0 then return true end
	local dif = (a.time or 0)/a.count - (b.time or 0)/b.count
	if dif ~= 0 then return dif > 0 end
	if (names[a] or a.name) ~= (names[b] or b.name) then return (names[a] or a.name) < (names[b] or b.name) end
	return (events[a] or a.event) < (events[b] or b.event)
end


function tekSupport_Events:OnTooltipUpdate(usetime)
	for event,ev in pairs(timetable) do
		for obj,v in pairs(ev) do
			if not insort[v] then
				insort[v] = true
				table.insert(sortedframes, v)
			end
			events[v] = events[v] or event
			names[v] = names[v] or type(obj) == "table" and (obj.name or obj.GetLibraryVersion and obj:GetLibraryVersion()) or tostring(obj)
		end
	end

	table.sort(sortedframes, usetime and SortTime or SortMem)

	local cat = tablet:AddCategory("columns", 6, "text", "Frame", "text2", "Event",
		"text3", "Mem", "text4", (not usetime and "|cffff0000" or "") .. "Mem/exec",
		"text5", "Time", "text6", (usetime and "|cffff0000" or "") .. "Time/exec",
		"child_justify3", "RIGHT", "child_justify4", "RIGHT", "child_justify5", "RIGHT")
	local i = 1
	for _,v in ipairs(sortedframes) do
		if i < maxttitems and v.mem and (((v.mem or 0)/(v.count or 1)) >= 1.0 or ((v.time or 0)/(v.count or 1)*1000) >= 1.0) then
			i = i + 1
			cat:AddLine("text", names[v] or v.name, "text2", (v.event or events[v]).. " x".. v.count,
				"text3", string.format("%d KiB", (v.mem or 0)),
				"text4", string.format("%.1f KiB", (v.mem or 0)/(v.count or 1)),
				"text5", string.format("%d ms", (v.time or 0)*1000),
				"text6", string.format("%.1f ms", (v.time or 0)/(v.count or 1)*1000))
		end
	end
end

